from pydantic import BaseModel

from gooey_gui import core
from gooey_gui.components import common as gui


class AlertDialogRef(BaseModel):
    key: str
    is_open: bool = False

    def set_open(self, value: bool):
        self.is_open = core.session_state[self.key] = value

    @property
    def open_btn_key(self):
        return self.key + ":open"

    @property
    def close_btn_key(self):
        return self.key + ":close"


class ConfirmDialogRef(AlertDialogRef):
    pressed_confirm: bool = False

    @property
    def confirm_btn_key(self):
        return self.key + ":confirm"


def use_confirm_dialog(
    key: str,
    close_on_confirm: bool = True,
) -> ConfirmDialogRef:
    ref = ConfirmDialogRef.parse_obj(use_alert_dialog(key))
    if not ref.is_open:
        return ref

    ref.pressed_confirm = bool(core.session_state.pop(ref.confirm_btn_key, None))
    if ref.pressed_confirm and close_on_confirm:
        ref.set_open(False)

    return ref


def use_alert_dialog(key: str) -> AlertDialogRef:
    ref = AlertDialogRef(key=key, is_open=bool(core.session_state.get(key)))
    if core.session_state.pop(ref.close_btn_key, None):
        ref.set_open(False)
    return ref


def alert_dialog(
    ref: AlertDialogRef,
    modal_title: str,
    large: bool = False,
) -> core.NestingCtx:
    header, body, _ = modal_scaffold(large=large)
    with header:
        gui.write(modal_title)
        gui.button(
            '<i class="fa fa-times fa-xl">',
            key=ref.close_btn_key,
            type="tertiary",
            className="m-0 py-1 px-2",
        )
    return body


def button_with_confirm_dialog(
    *,
    ref: ConfirmDialogRef,
    trigger_label: str,
    modal_title: str,
    modal_content: str | None = None,
    cancel_label: str = "Cancel",
    confirm_label: str,
    trigger_className: str = "",
    trigger_type: str = "secondary",
    cancel_className: str = "",
    confirm_className: str = "",
    large: bool = False,
) -> core.NestingCtx:
    if gui.button(
        label=trigger_label,
        key=ref.open_btn_key,
        className=trigger_className,
        type=trigger_type,
    ):
        ref.set_open(True)
    if ref.is_open:
        return confirm_dialog(
            ref=ref,
            modal_title=modal_title,
            modal_content=modal_content,
            cancel_label=cancel_label,
            confirm_label=confirm_label,
            cancel_className=cancel_className,
            confirm_className=confirm_className,
            large=large,
        )
    return gui.dummy()


def confirm_dialog(
    *,
    ref: ConfirmDialogRef,
    modal_title: str,
    modal_content: str | None = None,
    cancel_label: str = "Cancel",
    confirm_label: str,
    cancel_className: str = "",
    confirm_className: str = "",
    large: bool = False,
) -> core.NestingCtx:
    header, body, footer = modal_scaffold(large=large)
    with header:
        gui.write(modal_title)
    with footer:
        gui.button(
            label=cancel_label,
            key=ref.close_btn_key,
            className=cancel_className,
            type="tertiary",
        )
        gui.button(
            label=confirm_label,
            key=ref.confirm_btn_key,
            type="primary",
            className=confirm_className,
        )
    if modal_content:
        with body:
            gui.write(modal_content)
    return body


def modal_scaffold(
    large: bool = False,
) -> tuple[core.NestingCtx, core.NestingCtx, core.NestingCtx]:
    if large:
        large_cls = "modal-lg"
    else:
        large_cls = ""
    with core.current_root_ctx():
        with (
            gui.div(
                className="modal d-block",
                style=dict(zIndex="9999"),
                tabIndex="-1",
                role="dialog",
            ),
            gui.div(
                className=f"modal-dialog modal-dialog-centered {large_cls}",
                role="document",
            ),
            gui.div(className="modal-content border-0 shadow"),
        ):
            return (
                gui.div(className="modal-header border-0"),
                gui.div(className="modal-body"),
                gui.div(className="modal-footer border-0 py-0"),
            )



import asyncio
import logging
import sys
import traceback
from copy import copy
from pathlib import Path

import uvicorn
from uvicorn.server import HANDLED_SIGNALS
from uvicorn.supervisors import ChangeReload

from .pubsub import realtime_push, _extra_subscriptions

logger = logging.getLogger("uvicorn.error")


def runserver(*args, **kwargs):
    config = uvicorn.Config(*args, **kwargs)
    server = HotReloadServer(config)
    server.run()


class _UvicornServer(uvicorn.Server):
    async def startup(self, *args, **kwargs):
        ret = await super().startup(*args, **kwargs)
        if self.config.reload:
            realtime_push("--hot-reload", 1)
        return ret


class HotReloadServer:
    def __init__(self, config: uvicorn.Config):
        self.config = config
        self.server: _UvicornServer | None = None
        self.did_reload = False
        if self.config.reload:
            _extra_subscriptions.add("--hot-reload")

    def run(self, *args, **kwargs):
        self.config.setup_event_loop()
        return asyncio.run(self.serve(*args, **kwargs))

    async def serve(self, *args, **kwargs):
        watcher = ChangeReload(self.config, lambda _: None, [])
        task = asyncio.create_task(self.reloader(watcher))
        try:
            while True:
                config = copy(self.config)
                config.loaded = False
                self.server = _UvicornServer(config)
                self.did_reload = False
                try:
                    await self.server.serve(*args, **kwargs)
                except:
                    traceback.print_exc()
                    await self.wait_for_changes()
                if not self.did_reload:
                    return
                reload_modules(watcher)
        finally:
            logger.info("Stopping reloader")
            watcher.should_exit.set()
            task.cancel()

    async def wait_for_changes(self):
        should_exit = False

        def handle_exit():
            nonlocal should_exit
            should_exit = True

        loop = asyncio.get_event_loop()
        for sig in HANDLED_SIGNALS:
            loop.add_signal_handler(sig, handle_exit)

        while not (should_exit or self.did_reload):
            await asyncio.sleep(1)

    async def reloader(self, watcher):
        while True:
            loop = asyncio.get_event_loop()
            changes = await loop.run_in_executor(None, watcher.should_restart)
            if not changes:
                continue
            logger.warning(
                "%s detected changes in %s. Reloading...",
                watcher.reloader_name,
                ", ".join(map(str, changes)),
            )
            self.did_reload = True
            if self.server and self.server.started:
                self.server.should_exit = True


def reload_modules(watcher: ChangeReload):
    for name, module in list(sys.modules.items()):
        try:
            if not module.__file__:
                continue
        except AttributeError:
            continue
        modpath = Path(module.__file__)
        if any(
            modpath.match(pattern) for pattern in watcher.config.reload_excludes
        ) or not any(
            modpath.is_relative_to(directory) for directory in watcher.reload_dirs
        ):
            continue
        sys.modules.pop(name, None)



import inspect
import typing
from functools import partial, wraps

from fastapi import Depends
from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel
from starlette.requests import Request
from starlette.responses import JSONResponse, RedirectResponse, Response

from .exceptions import RedirectException, RerunException, StopException
from .pubsub import (
    get_subscriptions,
    realtime_clear_subs,
)
from .state import get_session_state, set_session_state, set_query_params, threadlocal

Style = dict[str, str | None]
ReactHTMLProps = dict[str, typing.Any]


def current_root_ctx() -> "NestingCtx":
    return threadlocal.root_ctx


class RenderTreeNode(BaseModel):
    name: str
    props: ReactHTMLProps = {}
    children: list["RenderTreeNode"] = []

    def mount(self) -> "RenderTreeNode":
        threadlocal.render_root.children.append(self)
        return self


class NestingCtx:
    def __init__(self, node: RenderTreeNode = None):
        self.node = node or threadlocal.render_root
        self.parent = None

    def __enter__(self):
        try:
            self.parent = threadlocal.render_root
        except AttributeError:
            pass
        threadlocal.render_root = self.node

    def __exit__(self, exc_type, exc_val, exc_tb):
        threadlocal.render_root = self.parent

    def empty(self):
        """Empty the children of the node"""
        self.node.children = []
        return self


async def request_json(request: Request) -> dict | None:
    if request.headers.get("content-type") == "application/json":
        return await request.json()


def route(app, *paths, **kwargs):
    def decorator(fn):
        @wraps(fn)
        def wrapper(request: Request, json_data: dict | None, **kwargs):
            if "request" in fn_sig.parameters:
                kwargs["request"] = request
            if "json_data" in fn_sig.parameters:
                kwargs["json_data"] = json_data
            return renderer(
                partial(fn, **kwargs),
                query_params=dict(request.query_params),
                state=json_data and json_data.get("state"),
            )

        fn_sig = inspect.signature(fn)
        mod_params = dict(fn_sig.parameters) | dict(
            request=inspect.Parameter(
                "request",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=Request,
            ),
            json_data=inspect.Parameter(
                "json_data",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=Depends(request_json),
                annotation=typing.Optional[dict],
            ),
        )
        mod_sig = fn_sig.replace(parameters=list(mod_params.values()))
        wrapper.__signature__ = mod_sig

        for path in reversed(paths):
            wrapper = app.get(path)(wrapper)
            wrapper = app.post(path)(wrapper)

        return wrapper

    return decorator


def renderer(
    render: typing.Callable,
    state: dict[str, typing.Any] = None,
    query_params: dict[str, str] = None,
) -> dict | Response:
    set_session_state(state or {})
    set_query_params(query_params or {})
    realtime_clear_subs()
    threadlocal.use_state_count = 0
    while True:
        try:
            root = RenderTreeNode(name="root")
            threadlocal.root_ctx = NestingCtx(root)
            with threadlocal.root_ctx:
                try:
                    ret = render()
                except StopException:
                    ret = None
                except RedirectException as e:
                    return RedirectResponse(e.url, status_code=e.status_code)
            if isinstance(ret, Response):
                return ret
            return JSONResponse(
                jsonable_encoder(
                    dict(
                        children=root.children,
                        state=get_session_state(),
                        channels=get_subscriptions(),
                        **(ret or {}),
                    )
                ),
                headers={"X-GOOEY-GUI-ROUTE": "1"},
            )
        except RerunException:
            continue



import base64
import html as html_lib
import math
import textwrap
import typing
from datetime import datetime, timezone

from furl import furl
from gooey_gui import core

T = typing.TypeVar("T")
LabelVisibility = typing.Literal["visible", "collapsed"]
TooltipPlacement = typing.Literal["left", "right", "top", "bottom", "auto"]

BLANK_OPTION = "———"


def _default_format(value: typing.Any) -> str:
    if value is None:
        return BLANK_OPTION
    return str(value)


def dummy(*args, **kwargs):
    return core.NestingCtx()


spinner = dummy
set_page_config = dummy
form = dummy
dataframe = dummy


def countdown_timer(
    end_time: datetime,
    delay_text: str,
) -> core.NestingCtx:
    return _node(
        "countdown-timer",
        endTime=end_time.astimezone(timezone.utc).isoformat(),
        delayText=delay_text,
    )


def nav_tabs():
    return _node("nav-tabs")


def nav_item(href: str, *, active: bool):
    return _node("nav-item", to=href, active="true" if active else None)


def nav_tab_content():
    return _node("nav-tab-content")


def div(**props) -> core.NestingCtx:
    return tag("div", **props)


def link(*, to: str, **props) -> core.NestingCtx:
    return _node("Link", to=to, **props)


def tag(tag_name: str, **props) -> core.NestingCtx:
    props["__reactjsxelement"] = tag_name
    return _node("tag", **props)


def html(body: str, **props):
    props["className"] = props.get("className", "") + " gui-html-container"
    return _node("html", body=body, **props)


def write(*objs: typing.Any, line_clamp: int = None, unsafe_allow_html=False, **props):
    for obj in objs:
        markdown(
            obj if isinstance(obj, str) else repr(obj),
            line_clamp=line_clamp,
            unsafe_allow_html=unsafe_allow_html,
            **props,
        )


def center(direction="flex-column", className="") -> core.NestingCtx:
    return div(
        className=f"d-flex justify-content-center align-items-center text-center {direction} {className}"
    )


def newline():
    html("<br/>")


def markdown(
    body: str | None,
    *,
    line_clamp: int = None,
    unsafe_allow_html=False,
    **props,
):
    if body is None:
        return _node("markdown", body="", **props)
    if not unsafe_allow_html:
        body = html_lib.escape(body)
    props["className"] = (
        props.get("className", "") + " gui-html-container gui-md-container"
    )
    return _node(
        "markdown",
        body=dedent(body).strip(),
        lineClamp=line_clamp,
        **props,
    )


def _node(nodename: str, **props):
    node = core.RenderTreeNode(name=nodename, props=props)
    node.mount()
    return core.NestingCtx(node)


def text(body: str, **props):
    core.RenderTreeNode(
        name="pre",
        props=dict(body=dedent(body), **props),
    ).mount()


def error(
    body: str,
    icon: str = "🔥",
    *,
    unsafe_allow_html=False,
    color="rgba(255, 108, 108, 0.2)",
    **props,
):
    if not isinstance(body, str):
        body = repr(body)
    with div(
        style=dict(
            backgroundColor=color,
            padding="1rem",
            paddingBottom="0",
            marginBottom="0.5rem",
            borderRadius="0.25rem",
            display="flex",
            gap="0.5rem",
        )
    ):
        markdown(icon)
        with div():
            markdown(dedent(body), unsafe_allow_html=unsafe_allow_html, **props)


def success(body: str, icon: str = "✅", *, unsafe_allow_html=False):
    if not isinstance(body, str):
        body = repr(body)
    with div(
        style=dict(
            backgroundColor="rgba(108, 255, 108, 0.2)",
            padding="1rem",
            paddingBottom="0",
            marginBottom="0.5rem",
            borderRadius="0.25rem",
            display="flex",
            gap="0.5rem",
        )
    ):
        markdown(icon)
        markdown(dedent(body), unsafe_allow_html=unsafe_allow_html)


def caption(body: str, className: str = None, **props):
    className = className or "text-muted"
    markdown(body, className=className, **props)


def tabs(labels: list[str]) -> list[core.NestingCtx]:
    parent = core.RenderTreeNode(
        name="tabs",
        children=[
            core.RenderTreeNode(
                name="tab",
                props=dict(label=dedent(label)),
            )
            for label in labels
        ],
    ).mount()
    return [core.NestingCtx(tab) for tab in parent.children]


def controllable_tabs(labels: list[str], key: str) -> tuple[list[core.NestingCtx], int]:
    index = core.session_state.get(key, 0)
    for i, label in enumerate(labels):
        if button(
            label,
            key=f"tab-{i}",
            type="primary",
            className="replicate-nav",
            style={
                "background": "black" if i == index else "white",
                "color": "white" if i == index else "black",
            },
        ):
            core.session_state[key] = index = i
            core.rerun()
    ctxs = []
    for i, label in enumerate(labels):
        if i == index:
            ctxs += [div(className="tab-content")]
        else:
            ctxs += [div(className="tab-content", style={"display": "none"})]
    return ctxs, index


def columns(
    spec,
    *,
    gap: str = None,
    responsive: bool = True,
    column_props: dict = {},
    **props,
) -> tuple[core.NestingCtx, ...]:
    if isinstance(spec, int):
        spec = [1] * spec
    total_weight = sum(spec)
    props.setdefault("className", "row")
    with div(**props):
        return tuple(
            div(
                className=f"col-lg-{p} {'col-12' if responsive else f'col-{p}'}",
                **column_props,
            )
            for w in spec
            if (p := f"{round(w / total_weight * 12)}")
        )


def image(
    src,
    caption: str = None,
    alt: str = None,
    href: str = None,
    show_download_button: bool = False,
    **props,
):
    try:
        import numpy as np

        is_np_img = isinstance(src, np.ndarray)
    except ImportError:
        is_np_img = False
    if is_np_img:
        if not src.shape:
            return
        # ensure image is not too large
        data = _export_cv2_img(src, 128)
        # convert to base64
        b64 = base64.b64encode(data).decode("utf-8")
        src = "data:image/png;base64," + b64
    if not src:
        return
    core.RenderTreeNode(
        name="img",
        props=dict(
            src=src,
            caption=dedent(caption),
            alt=alt or caption,
            href=href,
            **props,
        ),
    ).mount()
    if show_download_button:
        download_button(
            label='<i class="fa-regular fa-download"></i> Download', url=src
        )


def _export_cv2_img(img, size: int) -> bytes:
    """
    Resize the image to the given size, keeping the aspect ratio.
    """
    import cv2

    h, w = img.shape[:2]
    new_h = size
    new_w = int(w * new_h / h)
    img = cv2.resize(img, (new_w, new_h))
    data = cv2.imencode(".png", img)[1].tostring()
    return data


def video(
    src: str,
    caption: str = None,
    autoplay: bool = False,
    show_download_button: bool = False,
):
    autoplay_props = {}
    if autoplay:
        autoplay_props = {
            "preload": "auto",
            "controls": True,
            "autoPlay": True,
            "loop": True,
            "muted": True,
            "playsInline": True,
        }

    if not src:
        return
    if isinstance(src, str):
        # https://muffinman.io/blog/hack-for-ios-safari-to-display-html-video-thumbnail/
        f = furl(src)
        f.fragment.args["t"] = "0.001"
        src = f.url
    core.RenderTreeNode(
        name="video",
        props=dict(src=src, caption=dedent(caption), **autoplay_props),
    ).mount()
    if show_download_button:
        download_button(
            label='<i class="fa-regular fa-download"></i> Download', url=src
        )


def audio(src: str, caption: str = None, show_download_button: bool = False):
    if not src:
        return
    core.RenderTreeNode(
        name="audio",
        props=dict(src=src, caption=dedent(caption)),
    ).mount()
    if show_download_button:
        download_button(
            label='<i class="fa-regular fa-download"></i> Download', url=src
        )


def text_area(
    label: str,
    value: str = "",
    height: int = 500,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    placeholder: str = None,
    disabled: bool = False,
    label_visibility: LabelVisibility = "visible",
    **props,
) -> str:
    style = props.setdefault("style", {})
    # if key:
    #     assert not value, "only one of value or key can be provided"
    # else:
    if not key:
        key = core.md5_values(
            "textarea",
            label,
            height,
            help,
            placeholder,
            label_visibility,
            not disabled or value,
        )
    value = str(core.session_state.setdefault(key, value) or "")
    if label_visibility != "visible":
        label = None
    if disabled:
        max_height = f"{height}px"
        rows = nrows_for_text(value, height)
    else:
        max_height = "50vh"
        rows = nrows_for_text(value, height)
    style.setdefault("maxHeight", max_height)
    props.setdefault("rows", rows)
    core.RenderTreeNode(
        name="textarea",
        props=dict(
            name=key,
            label=dedent(label),
            defaultValue=value,
            help=help,
            tooltipPlacement=tooltip_placement,
            placeholder=placeholder,
            disabled=disabled,
            **props,
        ),
    ).mount()
    return value or ""


def code_editor(
    value: str = "",
    key: str = None,
    label: str = None,
    label_visibility: LabelVisibility = "visible",
    **props,
) -> typing.Any:
    value = str(core.session_state.setdefault(key, value) or "")
    if label_visibility != "visible":
        label = None
    core.RenderTreeNode(
        name="code-editor",
        props=dict(
            name=key,
            defaultValue=value,
            label=dedent(label),
            **props,
        ),
    ).mount()
    return value or ""


def nrows_for_text(
    text: str,
    max_height_px: int,
    min_rows: int = 1,
    row_height_px: int = 30,
    row_width_px: int = 70,
) -> int:
    max_rows = max_height_px // row_height_px
    nrows = math.ceil(
        sum(
            math.ceil(len(line) / row_width_px)
            for line in (text or "").splitlines(keepends=True)
        )
    )
    nrows = min(max(nrows, min_rows), max_rows)
    return nrows


def multiselect(
    label: str,
    options: typing.Sequence[T],
    format_func: typing.Callable[[T], typing.Any] = _default_format,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    allow_none: bool = False,
    *,
    disabled: bool = False,
) -> list[T]:
    if not options:
        return []
    options = list(options)
    if not key:
        key = core.md5_values("multiselect", label, options, help)
    value = core.session_state.get(key) or []
    if not isinstance(value, list):
        value = [value]
    value = [o for o in value if o in options]
    if not allow_none and not value:
        value = [options[0]]
    core.session_state[key] = value
    core.RenderTreeNode(
        name="select",
        props=dict(
            name=key,
            label=dedent(label),
            help=help,
            tooltipPlacement=tooltip_placement,
            isDisabled=disabled,
            isMulti=True,
            defaultValue=value,
            allow_none=allow_none,
            options=[
                {"value": option, "label": str(format_func(option))}
                for option in options
            ],
        ),
    ).mount()
    return value


def selectbox(
    label: str,
    options: typing.Iterable[T],
    format_func: typing.Callable[[T], typing.Any] = _default_format,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    *,
    disabled: bool = False,
    label_visibility: LabelVisibility = "visible",
    value: T = None,
    allow_none: bool = False,
    **props,
) -> T | None:
    if not options:
        return None
    if label_visibility != "visible":
        label = None
    options = list(options)
    if allow_none:
        options.insert(0, None)
    if not key:
        key = core.md5_values("select", label, options, help, label_visibility)
    value = core.session_state.setdefault(key, value)
    if value not in options:
        value = core.session_state[key] = options[0]
    core.RenderTreeNode(
        name="select",
        props=dict(
            name=key,
            label=dedent(label),
            help=help,
            tooltipPlacement=tooltip_placement,
            isDisabled=disabled,
            defaultValue=value,
            options=[
                {"value": option, "label": str(format_func(option))}
                for option in options
            ],
            **props,
        ),
    ).mount()
    return value


def download_button(
    label: str,
    url: str,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    *,
    type: typing.Literal["primary", "secondary", "tertiary", "link"] = "secondary",
    disabled: bool = False,
    **props,
) -> bool:
    url = furl(url).remove(fragment=True).url
    return button(
        component="download-button",
        url=url,
        label=label,
        key=key,
        help=help,
        tooltip_placement=tooltip_placement,
        type=type,
        disabled=disabled,
        **props,
    )


def button(
    label: str,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    *,
    type: typing.Literal["primary", "secondary", "tertiary", "link"] = "secondary",
    disabled: bool = False,
    component: typing.Literal["download-button", "gui-button"] = "gui-button",
    **props,
) -> bool:
    """
    Example:
        st.button("Primary", key="test0", type="primary")
        st.button("Secondary", key="test1")
        st.button("Tertiary", key="test3", type="tertiary")
        st.button("Link Button", key="test3", type="link")
    """
    if not key:
        key = core.md5_values("button", label, help, type, props)
    className = f"btn btn-theme btn-{type} " + props.pop("className", "")
    core.RenderTreeNode(
        name=component,
        props=dict(
            type="submit",
            value="yes",
            name=key,
            label=dedent(label),
            help=help,
            tooltipPlacement=tooltip_placement,
            disabled=disabled,
            className=className,
            **props,
        ),
    ).mount()
    return bool(core.session_state.pop(key, False))


def anchor(
    label: str,
    href: str,
    *,
    type: typing.Literal["primary", "secondary", "tertiary", "link"] = "secondary",
    disabled: bool = False,
    unsafe_allow_html: bool = False,
    new_tab: bool = False,
    **props,
):
    className = f"btn btn-theme btn-{type} " + props.pop("className", "")
    style = props.pop("style", {})
    if disabled:
        style["pointerEvents"] = "none"
    if new_tab:
        props["target"] = "_blank"
    with tag("a", href=href, className=className, style=style, **props):
        markdown(dedent(label), unsafe_allow_html=unsafe_allow_html)


form_submit_button = button


def expander(label: str, *, expanded: bool = False, key: str = None, **props):
    node = core.RenderTreeNode(
        name="expander",
        props=dict(
            label=dedent(label),
            open=expanded,
            name=key or core.md5_values(label, expanded, props),
            **props,
        ),
    )
    node.mount()
    return core.NestingCtx(node)


def file_uploader(
    label: str,
    accept: list[str] = None,
    accept_multiple_files=False,
    key: str = None,
    value: str | list[str] = None,
    upload_key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    *,
    disabled: bool = False,
    label_visibility: LabelVisibility = "visible",
    upload_meta: dict = None,
    optional: bool = False,
) -> str | list[str] | None:
    from .input_widgets import checkbox

    if label_visibility != "visible":
        label = None
    key = upload_key or key
    if not key:
        key = core.md5_values(
            "file_uploader",
            label,
            accept,
            accept_multiple_files,
            help,
            label_visibility,
        )
    if optional:
        if not checkbox(
            label,
            value=bool(core.session_state.get(key, value)),
            disabled=disabled,
            help=help,
        ):
            core.session_state.pop(key, None)
            return None
        label = None
    value = core.session_state.setdefault(key, value)
    if not value:
        if accept_multiple_files:
            value = []
        else:
            value = None
    core.session_state[key] = value
    core.RenderTreeNode(
        name="input",
        props=dict(
            type="file",
            name=key,
            label=dedent(label),
            help=help,
            tooltipPlacement=tooltip_placement,
            disabled=disabled,
            accept=accept,
            multiple=accept_multiple_files,
            defaultValue=value,
            uploadMeta=upload_meta,
        ),
    ).mount()
    return value


def json(value: typing.Any, expanded: bool = False, depth: int = 1, **props):
    core.RenderTreeNode(
        name="json",
        props=dict(
            value=value,
            expanded=expanded,
            defaultInspectDepth=3 if expanded else depth,
            **props,
        ),
    ).mount()


def data_table(file_url_or_cells: str | list):
    if isinstance(file_url_or_cells, str):
        file_url = file_url_or_cells
        return _node("data-table", fileUrl=file_url)
    else:
        cells = file_url_or_cells
        return _node("data-table-raw", cells=cells)


def table(df: "pd.DataFrame"):
    with tag("table", className="table table-striped table-sm"):
        with tag("thead"):
            with tag("tr"):
                for col in df.columns:
                    with tag("th", scope="col"):
                        html(dedent(col))
        with tag("tbody"):
            for row in df.itertuples(index=False):
                with tag("tr"):
                    for value in row:
                        with tag("td"):
                            html(dedent(str(value)))


def horizontal_radio(
    label: str,
    options: typing.Sequence[T],
    format_func: typing.Callable[[T], typing.Any] = _default_format,
    *,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    value: T = None,
    disabled: bool = False,
    checked_by_default: bool = True,
    label_visibility: LabelVisibility = "visible",
    **button_props,
) -> T | None:
    if not options:
        return None
    options = list(options)
    if not key:
        key = core.md5_values(
            "horizontal_radio", label, options, help, label_visibility
        )
    value = core.session_state.setdefault(key, value)
    if value not in options and checked_by_default:
        value = core.session_state[key] = options[0]
    if label_visibility != "visible":
        label = None
    markdown(label)
    for option in options:
        if button(
            format_func(option),
            key=f"tab-{key}-{option}",
            type="primary",
            className="replicate-nav " + ("active" if value == option else ""),
            disabled=disabled,
            **button_props,
        ):
            core.session_state[key] = value = option
            core.rerun()
    return value


def radio(
    label: str,
    options: typing.Sequence[T],
    format_func: typing.Callable[[T], typing.Any] = _default_format,
    key: str = None,
    value: T = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    *,
    disabled: bool = False,
    checked_by_default: bool = True,
    label_visibility: LabelVisibility = "visible",
) -> T | None:
    if not options:
        return None
    options = list(options)
    if not key:
        key = core.md5_values("radio", label, options, help, label_visibility)
    value = core.session_state.setdefault(key, value)
    if value not in options and checked_by_default:
        value = core.session_state[key] = options[0]
    if label_visibility != "visible":
        label = None
    markdown(label)
    for option in options:
        core.RenderTreeNode(
            name="input",
            props=dict(
                type="radio",
                name=key,
                label=dedent(str(format_func(option))),
                value=option,
                defaultChecked=bool(value == option),
                help=help,
                tooltipPlacement=tooltip_placement,
                disabled=disabled,
            ),
        ).mount()
    return value


def breadcrumbs(divider: str = "/", **props) -> core.NestingCtx:
    style = props.pop("style", {}) | {"--bs-breadcrumb-divider": f"'{divider}'"}
    with tag("nav", style=style, **props):
        return tag("ol", className="breadcrumb mb-0")


def breadcrumb_item(inner_html: str, link_to: str | None = None, **props):
    className = "breadcrumb-item " + props.pop("className", "")
    with tag("li", className=className, **props):
        if link_to:
            with tag("a", href=link_to):
                html(inner_html)
        else:
            html(inner_html)


def plotly_chart(figure_or_data, **kwargs):
    data = (
        figure_or_data.to_plotly_json()
        if hasattr(figure_or_data, "to_plotly_json")
        else figure_or_data
    )
    core.RenderTreeNode(
        name="plotly-chart",
        props=dict(
            chart=data,
            args=kwargs,
        ),
    ).mount()


def popover(**props) -> tuple[core.NestingCtx, core.NestingCtx]:
    content = core.RenderTreeNode(name="content")

    popover = core.RenderTreeNode(
        name="popover", props=props | dict(content=content.children)
    )
    popover.mount()

    return core.NestingCtx(popover), core.NestingCtx(content)


def dedent(text: str | None) -> str | None:
    if not text:
        return text
    return textwrap.dedent(text)


def js(src: str, **kwargs):
    core.RenderTreeNode(
        name="script",
        props=dict(
            src=src,
            args=kwargs,
        ),
    ).mount()


def switch(
    label: str,
    value: bool = False,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    *,
    disabled: bool = False,
    size: typing.Literal["small", "large"] = "large",
    label_visibility: LabelVisibility = "visible",
    **props,
) -> bool:
    if not key:
        key = core.md5_values("input", "switch", label, help, label_visibility)
    value = core.session_state.setdefault(key, value)
    if label_visibility != "visible":
        label = None
    core.RenderTreeNode(
        name="switch",
        props={
            "name": key,
            "label": dedent(label),
            "defaultChecked": value,
            "help": help,
            "tooltipPlacement": tooltip_placement,
            "disabled": disabled,
            "size": size,
            **props,
        },
    ).mount()
    return bool(value)


def tooltip(
    content: str = "",
    *,
    placement: TooltipPlacement | None = None,
    **props,
) -> core.NestingCtx:
    tooltip = core.RenderTreeNode(
        name="tooltip",
        props={
            "content": content,
            "placement": placement,
            **props,
        },
    )
    tooltip.mount()
    return core.NestingCtx(tooltip)



from .common import *
from .modal import (
    AlertDialogRef,
    ConfirmDialogRef,
    use_alert_dialog,
    modal_scaffold,
    alert_dialog,
    use_confirm_dialog,
    confirm_dialog,
    button_with_confirm_dialog,
)
from .input_widgets import (
    text_input,
    date_input,
    number_input,
    password_input,
    slider,
    checkbox,
)
from .pills import pill
from .url_button import url_button



import urllib.parse


class RedirectException(Exception):
    def __init__(self, url, status_code=302):
        self.url = url
        self.status_code = status_code


class QueryParamsRedirectException(RedirectException):
    def __init__(self, query_params: dict, status_code=303):
        query_params = {k: v for k, v in query_params.items() if v is not None}
        url = "?" + urllib.parse.urlencode(query_params)
        super().__init__(url, status_code)


class StopException(Exception):
    pass


class RerunException(Exception):
    pass


def rerun():
    raise RerunException()


def stop():
    raise StopException()



import base64
import html as html_lib
import math
import textwrap
import typing
from datetime import datetime, timezone

from furl import furl
from gooey_gui import core
from loguru import logger

from .common import LabelVisibility, TooltipPlacement, dedent


def text_input(
    label: str,
    value: str = "",
    max_chars: str = None,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    *,
    placeholder: str = None,
    disabled: bool = False,
    label_visibility: LabelVisibility = "visible",
    **props,
) -> str:
    value = _html_input(
        "text",
        label=label,
        value=value,
        key=key,
        help=help,
        tooltip_placement=tooltip_placement,
        disabled=disabled,
        label_visibility=label_visibility,
        maxLength=max_chars,
        placeholder=placeholder,
        **props,
    )
    return value or ""


def date_input(
    label: str,
    value: str | None = None,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    *,
    disabled: bool = False,
    label_visibility: LabelVisibility = "visible",
    **props,
) -> datetime | None:
    value = _html_input(
        "date",
        label=label,
        value=value,
        key=key,
        help=help,
        tooltip_placement=tooltip_placement,
        disabled=disabled,
        label_visibility=label_visibility,
        style=dict(
            border="1px solid hsl(0, 0%, 80%)",
            padding="0.375rem 0.75rem",
            borderRadius="0.25rem",
            margin="0 0.5rem 0 0.5rem",
        ),
        **props,
    )
    try:
        return datetime.strptime(value, "%Y-%m-%d") if value else None
    except ValueError:
        return None


def password_input(
    label: str,
    value: str = "",
    max_chars: str = None,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    *,
    placeholder: str = None,
    disabled: bool = False,
    label_visibility: LabelVisibility = "visible",
    **props,
) -> str:
    value = _html_input(
        "password",
        label=label,
        value=value,
        key=key,
        help=help,
        tooltip_placement=tooltip_placement,
        disabled=disabled,
        label_visibility=label_visibility,
        maxLength=max_chars,
        placeholder=placeholder,
        **props,
    )
    return value or ""


def slider(
    label: str,
    min_value: float = None,
    max_value: float = None,
    value: float = None,
    step: float = None,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    *,
    disabled: bool = False,
) -> float:
    value = _html_input(
        "range",
        label=label,
        value=value,
        key=key,
        help=help,
        tooltip_placement=tooltip_placement,
        disabled=disabled,
        min=min_value,
        max=max_value,
        step=_step_value(min_value, max_value, step),
    )
    return value or 0


def number_input(
    label: str,
    min_value: float = None,
    max_value: float = None,
    value: float = None,
    step: float = None,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    *,
    disabled: bool = False,
) -> float:
    value = _html_input(
        "number",
        inputMode="decimal",
        label=label,
        value=value,
        key=key,
        help=help,
        tooltip_placement=tooltip_placement,
        disabled=disabled,
        min=min_value,
        max=max_value,
        step=_step_value(min_value, max_value, step),
    )
    return value or 0


def _step_value(
    min_value: float | None, max_value: float | None, step: float | None
) -> float:
    if step:
        return step
    elif isinstance(min_value, float) or isinstance(max_value, float):
        return 0.1
    else:
        return 1


def checkbox(
    label: str,
    value: bool = False,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    *,
    disabled: bool = False,
    label_visibility: LabelVisibility = "visible",
    **props,
) -> bool:
    value = _html_input(
        "checkbox",
        label=label,
        value=value,
        key=key,
        help=help,
        tooltip_placement=tooltip_placement,
        disabled=disabled,
        label_visibility=label_visibility,
        default_value_attr="defaultChecked",
        **props,
    )
    return bool(value)


def _html_input(
    input_type: str,
    /,
    *,
    label: str,
    value: typing.Any = None,
    key: str = None,
    help: str | None = None,
    tooltip_placement: TooltipPlacement | None = None,
    disabled: bool = False,
    label_visibility: LabelVisibility = "visible",
    default_value_attr: str = "defaultValue",
    **kwargs,
) -> typing.Any:
    if not key:
        key = core.md5_values("input", input_type, label, help, label_visibility)
    value = core.session_state.setdefault(key, value)
    if label_visibility != "visible":
        label = None
    core.RenderTreeNode(
        name="input",
        props={
            "type": input_type,
            "name": key,
            "label": dedent(label),
            default_value_attr: value,
            "help": help,
            "tooltipPlacement": tooltip_placement,
            "disabled": disabled,
            **kwargs,
        },
    ).mount()
    return value



import html
from typing import Literal

from gooey_gui.components import common as gui


def pill(
    title: str,
    *,
    unsafe_allow_html: bool = False,
    text_bg: Literal["primary", "secondary", "light", "dark", None] = "light",
    className: str = "",
    **props,
):
    if not unsafe_allow_html:
        title = html.escape(title)

    className += f" badge rounded-pill"
    if text_bg:
        className += f" text-bg-{text_bg}"

    with gui.tag("span", className=className):
        gui.html(title, **props)
